package wordcheck;

import eu.rekawek.toxiproxy.ToxiproxyClient;
import eu.rekawek.toxiproxy.model.ToxicDirection;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockserver.verify.VerificationTimes;
import org.testcontainers.containers.MockServerContainer;
import org.testcontainers.containers.Network;
import org.testcontainers.containers.ToxiproxyContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;
import org.testcontainers.utility.DockerImageName;
import org.mockserver.client.MockServerClient;

import java.io.IOException;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockserver.model.HttpRequest.request;
import static org.mockserver.model.HttpResponse.response;

@Testcontainers
public class AgileDeveloperSpellCheckTest {
  private AgileDeveloperSpellCheck agileDeveloperSpellCheck;

  private static final Network network = Network.newNetwork();
  private static final String NETWORK_ALIAS = "spellchecker";

  @Container
  private static final MockServerContainer mockServerContainer =
    new MockServerContainer(DockerImageName.parse("mockserver/mockserver"))
      .withNetwork(network)
      .withNetworkAliases(NETWORK_ALIAS);

  @Container
  private static final ToxiproxyContainer toxiProxyContainer =
    new ToxiproxyContainer("ghcr.io/shopify/toxiproxy:2.5.0")
      .withNetwork(network);

  private static MockServerClient mockServerClient;

  @BeforeAll
  static void beforeAll() {
    mockServerClient = new MockServerClient(
      mockServerContainer.getHost(), mockServerContainer.getServerPort());
  }

  @BeforeEach
  void init() {
    mockServerClient.reset();

    agileDeveloperSpellCheck = new AgileDeveloperSpellCheck();
  }

  private void runTestWithMockServerClient(String word, boolean expected) {
    mockServerClient
      .when(request()
        .withPath("/spellcheck")
        .withQueryStringParameter("check", word))
      .respond(
        response().withBody(Boolean.toString(expected)));

    var url = mockServerContainer.getEndpoint();
    System.setProperty("SPELLCHECKER_URL", url);

    assertEquals(expected, agileDeveloperSpellCheck.isSpellingCorrect(word));

    mockServerClient.verify(
      request()
        .withMethod("GET")
        .withPath("/spellcheck")
        .withQueryStringParameter("check", word),
      VerificationTimes.once());
  }

  @Test
  void checkSpellingForACorrectWord(){
    runTestWithMockServerClient("good", true);
  }

  @Test
  void checkSpellingForAnotherCorrectWord(){
    runTestWithMockServerClient("nice", true);
  }

  @Test
  void checkSpellingForAIncorrectWord(){
    runTestWithMockServerClient("aer", false);
  }

  @Test
  void checkSpellingForAnotherIncorrectWord(){
    runTestWithMockServerClient("hwo", false);
  }

  @Test
  void checkSpellingThrowsAnExceptionUponNetworkFailure() throws IOException {
    var toxiProxyClient = new ToxiproxyClient(
      toxiProxyContainer.getHost(), toxiProxyContainer.getControlPort());

    var proxy = toxiProxyClient.createProxy(
      NETWORK_ALIAS,
      "0.0.0.0:8666",
      "%s:%d".formatted(NETWORK_ALIAS, mockServerContainer.getFirstMappedPort()));

    var url = "http://%s:%d".formatted(toxiProxyContainer.getHost(), toxiProxyContainer.getMappedPort(8666));
    System.setProperty("SPELLCHECKER_URL", url);

    proxy.toxics().bandwidth("CUT_CONNECTION_DOWNSTREAM", ToxicDirection.DOWNSTREAM, 0);
    proxy.toxics().bandwidth("CUT_CONNECTION_UPSTREAM", ToxicDirection.UPSTREAM, 0);

    var ex = assertThrows(RuntimeException.class,
      () -> agileDeveloperSpellCheck.isSpellingCorrect("fiasco"));

    assertInstanceOf(IOException.class, ex.getCause());

    proxy.toxics().get("CUT_CONNECTION_DOWNSTREAM").remove();
    proxy.toxics().get("CUT_CONNECTION_UPSTREAM").remove();
  }
}
